#@ File (label = "image" ) imgpath
#@ File (label = "output folder", style = "directory")  outdir_spotfinder
#@ File (label = "output summary file", style = "file")  outfile_summary
#@ String (label = "Projection type" , choices = {"sum", "max"}) zproj
#@ Float (label = "LoG sigma (1-2)", value = 1) sigma
#@ Float (label = "Prominence Find Maxima (0.5-1)", value = 0.5) prominence
#@ Float (label = "Threshold factor Find Maxima (1-2)", value = 1) thrfactor
#@ Integer (label = "Dilation spots for nuclei mask (5-15)", value = 7) spot_dilation
#@ Integer  (label = "Threshold nuclei mask ", value = 11) thr_mask
#@ Integer (label = "Min area nuclei mask prior shrink (um2) (20-100)", value = 20) min_area_mask
#@ Integer  (label = "Shrink nuclei mask to compensate dilation (10-20)", value = 20) shrink_mask
#@ Integer  (label = "Refine diameter centerofmass (5-10)", value = 5) refine_diameter

#@ Output Float densityMask
#@ Output Float densityBestMask
#@ Output Float densityConvex
#@ Output Float densityBestConvex
#@ Output Float areaMask
#@ Output Float areaConvex
#@ Output Integer bestSlice
#@ Output Float varBestSlice
#@ Output Float varSumSlices
#@ Output Float IntSumSlices

import com.ibm.icu.util.Measure
import ij.*
import ij.measure.Measurements
import ij.plugin.ZProjector as ZProjector
import ij.plugin.frame.RoiManager as RoiManager
import ij.gui.PointRoi as PointRoi
import ij.process.ImageProcessor
import ij.process.ImageStatistics
import imagescience.feature.Laplacian as Laplacian
import imagescience.image.Image as Image
import com.opencsv.CSVWriter
import java.util.Collections;
import ij.gui.OvalRoi;
import org.apache.commons.io.FilenameUtils
import ij.plugin.Duplicator;
import ij.gui.Roi;
// Antonio Politi
// apoliti@mpibpc.mpg.de MPIBPC April 2021
// Script to find small spots, e.g. NPCs
// register image stack 
// filter by LoG
// perform spot finding (find maxima) on sum-projected laplacian image
// Select spots according to threshold in laplace and raw image

// TODO: Add again the alignment option
// TODO: Calibration of output image
 // Initial image is already single cell
version = "0.1.4";
show = true
main(outdir_spotfinder, imgpath, zproj, sigma, prominence, thrfactor,
		spot_dilation, thr_mask, min_area_mask,
		shrink_mask, refine_diameter);

def assignDefaultOutput(){
	densityMask = 0
	densityBestMask = 0
	densityConvex = 0
	densityBestConvex = 0
	areaMask = 0
	areaConvex = 0
	bestSlice = 0
	varBestSlice = 0
	varSumSlices = 0
	IntSumSlices = 0

}


def exportOutput(outfile){
	header = new String[]{"FullPath", "ParentPath", "Filename", "densityMask_um2", "densityBestMask_um2", "densityConvex_um2",
			"densityBestConvex_um2", "areaMask_um2", "areaConvex_um2", "bestSlice", "varBestSlice",
	"varSumSlices", "IntSumSlices"}
	final List<String[]> dataout = new ArrayList<>();

	dataout.add( new String[]{ imgpath.getPath(), imgpath.getParent(), imgpath.getName(),  "" + densityMask , "" + densityBestMask ,
			"" + 	densityConvex , "" +	densityBestConvex,
			"" + areaMask, "" + areaConvex,  "" +  bestSlice, "" + varBestSlice,
			"" +  varSumSlices, "" +  IntSumSlices}
)
	writeDataAtOnce(outfile, header, dataout, true);
}



def main(File outdir_loc, File imgpath_loc, String proj_z, Float sigma_log, Float prominence_findmaxima,
		 Float thrfactor_findmaxima, Integer spot_dilation_nucleimask, Integer thr_nucleimask,
		Integer min_area_nucleimask, Integer shrink_nucleimask, Integer refine_diameter_coord ) {
	// Initialization

	IJ.run("Close All")
	def outfiles = getFilenames(outdir_loc.getPath(), FilenameUtils.removeExtension(imgpath_loc.getName()), proj_z)

	// Save parameters of workflow
	File outputfilepar = new File(outfiles["parcsv"])
	saveParCsv(outputfilepar, version)

	// get Roimanager
	def rm = getRoiManager()

	// open the image


	def imp = openImage(imgpath_loc.getPath())
	if (show) {
		imp.show()
	}
	def calImg = imp.getCalibration()
	pixelSize = [calImg.pixelWidth,
				 calImg.pixelHeight]

	min_area_nucleimask = min_area_nucleimask/(pixelSize[0]*pixelSize[1])
	// Register the stack no interpolation
	def imp_aligned = register(imp, "SIFT");
	imp_aligned.setTitle("registered");


	// Zprojected image
	def imp_Zproj = ZProjector.run(imp_aligned, proj_z)

	// Smoothed sum image
	def imp_smooth = imp_aligned.duplicate()
	IJ.run(imp_smooth, "Smooth", "stack")
	def imp_smooth_Zproj = ZProjector.run(imp_smooth, proj_z)
	if (show) {		imp_smooth_Zproj.show() }

	// Variance image of smoothed
	def imp_smooth_Zproj_var = imp_smooth_Zproj.duplicate()
	IJ.run(imp_smooth_Zproj_var, "Variance...", "radius=2");


	// LoG filtering
	def imp_laplace = laplace(imp_aligned, sigma_log);
	imp_laplace.setTitle("registered_laplace");
	def imp_laplace_Zproj = ZProjector.run(imp_laplace, proj_z);
	if (show)  {		imp_laplace_Zproj.show();}


	// Find spots and filter them
	def rois_spots = runFindMaxima(imp_laplace_Zproj, imp_smooth_Zproj, prominence_findmaxima, "Otsu", thrfactor_findmaxima);

	// Create mask
	def roiMask = createMaskFromPores(imp_smooth_Zproj_var, rm, thr_nucleimask, spot_dilation_nucleimask, min_area_nucleimask, shrink_nucleimask);
	if (!roiMask[0]) {
		assignDefaultOutput()
		return
	}
	// Filter pores using the mask
	maskPores(rois_spots, roiMask, rm, imp);


	rois_spots["points_filter_masked_refined"] = refinePosition(rois_spots["points_filter_masked"], imp_smooth_Zproj, refine_diameter_coord );


	// best slice is base
	best_slice = getSlicewithMaxContrast(imp_aligned, rois_spots["mask"], 2);


	def imp_smooth_bestslice = (new Duplicator()).run(imp_smooth, (int) best_slice[0], (int) best_slice[0])
	def imp_laplace_bestslice = (new Duplicator()).run(imp_laplace, (int) best_slice[0], (int) best_slice[0])
	if (show) {	
		imp_laplace_bestslice.show()
		imp_smooth_bestslice.show()
	
	}
	def rois_spots_bestslice = runFindMaxima(imp_laplace_bestslice, imp_smooth_bestslice, prominence_findmaxima, "Otsu", thrfactor_findmaxima);
	rois_spots_bestslice = maskPores(rois_spots_bestslice, roiMask, rm, imp);
	rois_spots_bestslice["points_filter_masked_refined"] = refinePosition(rois_spots_bestslice["points_filter_masked"], imp_smooth_bestslice, refine_diameter_coord );
	// Add rois
	for (roi in rois_spots) {
		rm.addRoi(roi.value);
	}
	rois_spots_bestslice["points_filter_masked"].setName("points_filter_masked_BS")
	rois_spots_bestslice["points_filter_masked"].setPosition((int) best_slice[0])
	rm.addRoi(rois_spots_bestslice["points_filter_masked"])
	rois_spots_bestslice["points_filter_masked_refined"].setName("points_filter_masked_refined_BS")
	rois_spots_bestslice["points_filter_masked_refined"].setPosition((int) best_slice[0])
	rm.addRoi(rois_spots_bestslice["points_filter_masked_refined"])


	// Create output
	imp.setRoi(rois_spots["convex_area"]);
	areaConvex = imp.getStatistics().area;
	if (roiMask[0]) {
		imp.setRoi(rois_spots["mask"]);
		areaMask = imp.getStatistics().area;
	} else {
		areaMask = 1;
	}
	
	rm.runCommand("save", outfiles["roizip"]);
	File outputfile = new File(outfiles['resultcsv']);
	File outputfileBS = new File(outfiles['resultBScsv']);
	nrspots = savePointsCsv(rois_spots['points_filter_masked'], [imp_laplace_Zproj.getProcessor(), imp_smooth_Zproj.getProcessor()],
			pixelSize, areaConvex, areaMask, -1, outputfile);


	nrspotsBest = savePointsCsv(rois_spots_bestslice['points_filter_masked'], [imp_laplace_bestslice.getProcessor(), imp_smooth_bestslice.getProcessor()],
			pixelSize, areaConvex, areaMask, (int) best_slice[0], outputfileBS);
	densityConvex = nrspots/areaConvex;
	densityMask = nrspots/areaMask;
	densityBestConvex = nrspotsBest/areaConvex;
	densityBestMask = nrspotsBest/areaMask;
	bestSlice = (int) best_slice[0]
	varBestSlice = best_slice[1]
	varSumSlices = best_slice[2]
	IntSumSlices = best_slice[3]
	exportOutput(outfile_summary)


	// Create last output 
	rm.reset()
	imp_smooth_Zproj.setCalibration(calImg)
	imp_aligned.setCalibration(calImg)
	imp_smooth_Zproj.deleteRoi()
	IJ.run(imp_smooth_Zproj, "Enhance Contrast", "saturated=0.35")
	rm.addRoi(rois_spots['mask'])
	rois_spots['points_filter_masked_refined'].setPointType(PointRoi.CIRCLE)

	rm.addRoi(rois_spots['points_filter_masked_refined'])
	rm.runCommand(imp_smooth_Zproj,"Show All without labels");
	impf = imp_smooth_Zproj.flatten()
	impf.show()

	IJ.saveAs(imp_smooth_Zproj, "Tiff", outfiles["alignedprojtif"])
	IJ.saveAs(imp_aligned, "Tiff", outfiles["alignedtif"]);
	IJ.saveAs(impf, "jpeg", outfiles["alignedprojjpeg"])
}

/**
 *
 * @param imp1
 * @param imp2
 * @param thr_method
 * @param prominence
 * @param thrfactor
 * @return
 */
def runFindMaxima(ImagePlus imp1, ImagePlus imp2, Double prominence, String thr_method,  Double thrfactor) {

	IJ.run(imp1, "Find Maxima...", "prominence=" + prominence + " output=[Point Selection]")
	points = imp1.getRoi()
	points.setName("points")

	// Filter spots on intensity of laplace Zproj and smooth
	def ips = [imp1.getProcessor(), imp2.getProcessor()]
	ips[0].setAutoThreshold(thr_method)
	ips[1].setAutoThreshold(thr_method)
	thr = [ips[0].getMaxThreshold() * thrfactor, ips[1].getMaxThreshold() * thrfactor]

	points_filter = filter_maxima(points, ips, thr)
	points_filter.setName("points_filter")

	return [points: points, points_filter: points_filter]

}
/**
 *
 * @param imp The intensity image
 * @param mask Roi
 * @param var_radius Radius of Variance filter
 * @return [Slice where the variance is maximal, value of variance]
 */
def getSlicewithMaxContrast(ImagePlus imp, mask, var_radius){

	imp_var = imp.duplicate();
	IJ.run(imp_var, "Variance...", "radius=" + var_radius + " stack");
	def var_values = [];
	var_tot = 0;
	int_tot = 0;
	for (islice = 0; islice < imp.getNSlices(); islice++) {
		imp_var.setSlice(islice);
		imp.setSlice(islice);
		imp_var.setRoi(mask);
		imp.setRoi(mask);
		stat_var = imp_var.getStatistics();
		stat = imp.getStatistics();
		var_tot += stat_var.mean;
		int_tot += stat.mean;
		var_values.add(stat_var.mean);


	}
	idx_max = var_values.findIndexValues {
		it == var_values.max();
	}[0]
	return [idx_max+1, var_values[idx_max], var_tot, int_tot]
}

def getFilenames(outdir, imgname, proj) {
	return ["roizip" : FilenameUtils.concat(outdir, imgname + ".zip"),
			"resultcsv":  FilenameUtils.concat(outdir, imgname + ".csv"),
			"resultBScsv":  FilenameUtils.concat(outdir, imgname + "_BS.csv"),
			"parcsv" :  FilenameUtils.concat(outdir, imgname + "_par.csv"),
			"alignedtif" : FilenameUtils.concat(outdir, imgname + "_Aligned.tif"),
			"alignedprojtif" : FilenameUtils.concat(outdir, imgname + "_Aligned_"+ proj + "Proj_Smooth.tif"),
			"alignedprojjpeg" : FilenameUtils.concat(outdir, imgname + "_Aligned_"+ proj + "Proj_Smooth.jpg")]
}

def getRoiManager() {
	RoiManager rm;
	rm = new RoiManager();
	rm = RoiManager.getInstance()
	if (rm.getCount() > 0) {
		rm.reset();
	}
	return rm;
}


def openImage(imgname) {
	imp = IJ.openImage(imgpath.toString());
	imp.show()
	try {
		IJ.run(imp, "signed16bit to unsigned", "");
	} catch (Exception e) {
		IJ.showMessage("Please enable the liveim-tools update site! [Help > Update > Manage update site > [x] liveim-tools]")
	}
	
	imp.setTitle("raw");

	imp.hide()
	return imp;
}




def maskPores(spots, mask, rm, imp){
	rm.reset();
	if (!mask[0]) {
		IJ.log("no Nuclei Roi found use the spots");
		imp.setRoi(spots[1]);
		IJ.run(imp, "Convex Hull", "");
		roiConvex = imp.getRoi();
		roiConvex.setName("convex_area");
		spots.add(roiConvex);
		return spots;
	} else {
		spots['mask'] = mask[1];
		spots.each { entry -> rm.addRoi(entry.value)};
		rm.setSelectedIndexes(new int[]{1, 2});
		rm.runCommand(imp, "AND");
		points_filter_masked = imp.getRoi();
		points_filter_masked.setName("points_filter_masked");
		IJ.run(imp, "Convex Hull", "");
		roiConvex = imp.getRoi();
		roiConvex.setName("convex_area");
		spots["points_filter_masked"] = points_filter_masked;
		spots["convex_area"] =  roiConvex;
		rm.reset();
	}
	return spots;
}










// A second step is to perform an and operation with mask of nuclei this is a separate script

def getCenterRoi(ImagePlus imp, Roi[] rois) {
	def w = imp.getWidth()
	def h = imp.getHeight()
	List <Double> distance = new ArrayList<>();
	for (roi in rois) {
		imp.setRoi(roi);
		distance.add((w/2.0 - imp.getStatistics(Measurements.CENTROID).xCentroid)**2 + (h/2.0 - imp.getStatistics(Measurements.CENTROID).yCentroid)**2);
	}
	int ind_dist = distance.indexOf(Collections.min(distance));
	return rois[ind_dist];
}

def getMaxAreaRoi(ImagePlus imp, Roi[] rois) {
	List <Double> area = new ArrayList<>();
	for (roi in rois){
		area.add(roi.getStatistics().area);
	}
	// Get the maximal area
	int ind_area = area.indexOf(Collections.max(area));
	return rois[ind_area]
}

def createMaskFromPores(ImagePlus imp, RoiManager rm, thr, opening_radius, nuclei_area, shrink_mask) {
	IJ.run(imp, "Morphological Filters", "operation=Dilation element=Disk radius=" + opening_radius);
	imp_dilated = IJ.getImage();
	maxValue = imp_dilated.getProcessor().getMax();
	imp_dilated.getProcessor().setThreshold(thr, maxValue, 1);
	IJ.run(imp_dilated, "Convert to Mask", "");
	IJ.run(imp_dilated, "Analyze Particles...", "size="+nuclei_area+"-Infinity display include add");
	if (rm.getCount() == 0)
		return [false, 0];

	// local variable
	def rois = rm.getRoisAsArray();
	roi = getCenterRoi(imp, rois);
	rm.reset();
	imp.setRoi(roi);
	rm.addRoi(roi);
	rm.select(0);
	IJ.run(imp, "Enlarge...", "enlarge=-" + shrink_mask + " pixel");

	rm.runCommand(imp,"Update");
	// Perform a second particle analysis tro remove too small of objects keep only the largest area
	maskip = imp.createRoiMask()
	maskimp = new ImagePlus("mask_eroded", maskip)
	rm.reset()
	IJ.run(maskimp, "Analyze Particles...", "size="+nuclei_area+"-Infinity display include add");
	rois = rm.getRoisAsArray();
	if (rm.getCount() == 0)
		return [false, 0];
	rois_out = getMaxAreaRoi(imp, rois);
	rois_out.setName("mask");
	rm.reset();
	return [true, rois_out];
}



def register(imp, method) {
	// return a registered image or the same image depending on the method
	if (method == "SIFT"){
		imp.show();
		IJ.run("Linear Stack Alignment with SIFT",
				"initial_gaussian_blur=1.60 steps_per_scale_octave=3 " +
						"minimum_image_size=64 maximum_image_size=1024 feature_descriptor_size=4 " +
						"feature_descriptor_orientation_bins=8 closest/next_closest_ratio=0.92 " +
						"maximal_alignment_error=10 inlier_ratio=0.05 expected_transformation=Rigid");
		imp_aligned = IJ.getImage();
		return imp_aligned;
	} else {
		imp_aligned = imp.duplicate();
		return imp;
	}
}

def laplace(imp, sigma){
	// Compute LoG of image
	
	//IJ.run("FeatureJ Laplacian", "compute smoothing=1"); // 1 or 1.5 do not pick something too high
	//lap = Class.forName("imagescience.feature.Laplacian").newInstance()

	lap = new Laplacian();
	// This is a 3D laplacian!
	imp_laplace = lap.run(Image.wrap(imp), sigma);
	imp_laplace = imp_laplace.imageplus();
	IJ.run(imp_laplace, "Multiply...", "value=-1 stack");
	return imp_laplace
}

def filter_maxima(PointRoi points, ip, thr) {
	// filter maxima according to some intensity thresholds
	PTs = points.getContainedFloatPoints();
	points_clean = new PointRoi();
	def pixValue = [0, 0];
	for (int ipt = 0; ipt < PTs.npoints; ipt++){
		pixValue[0] = ip[0].getPixelValue((int) PTs.xpoints[ipt], (int) PTs.ypoints[ipt]);
		pixValue[1] = ip[1].getPixelValue((int) PTs.xpoints[ipt], (int) PTs.ypoints[ipt]);
	
		if ( (pixValue[0] > thr[0]) && (pixValue[1] > thr[1])) {
			points_clean.addPoint( PTs.xpoints[ipt],  PTs.ypoints[ipt]);
		}
	
	}
	return points_clean
}


def refinePosition(PointRoi points,  ImagePlus imp, roidiameter){
	PTs = points.getContainedFloatPoints();
	points_com = new PointRoi();
	points_com.setName("points_filter_masked_refined");
	
	for (int ipt = 0; ipt < PTs.npoints; ipt++){
		imp.setRoi(new OvalRoi(PTs.xpoints[ipt]-roidiameter/2, PTs.ypoints[ipt]-roidiameter/2, roidiameter, roidiameter))
		statCoord = imp.getStatistics(Measurements.CENTROID + Measurements.CENTER_OF_MASS)

		points_com.addPoint(statCoord.xCenterOfMass, statCoord.yCenterOfMass)
	}

	return points_com
}

def saveParCsv(outfile, version){
	final List<String[]> dataout = new ArrayList<>();
	
	dataout.add(new String[]{ imgpath.getPath(), outdir_spotfinder.getPath() , outfile_summary.getPath(),  zproj, "" + sigma ,   "SIFT", "" + prominence,
				"" + "Otsu", "" + thrfactor, "" + spot_dilation, "" + thr_mask, ""+ min_area_mask,
			"" + shrink_mask, "" + refine_diameter, "" + version})
	

	header = new String[]{"imgpath", "outdir_spotfinder", "outfile_summary", "zproj", "sigma", "reg_method", "prominence", "thr_method", "thrfactor",
	"spot_dilation", "thr_mask", "min_area_mask", "shrink_mask", "refine_diamter", "version"};
	writeDataAtOnce(outfile, header, dataout, false);

}

def savePointsCsv(PointRoi points,  ip, pixelSize, area_convex, area_mask, bestslice, outfile) {

	PTs = points.getContainedPoints();
	final List<String[]> dataout = new ArrayList<>();

	def pixValue = [0, 0];
	for (int ipt = 0; ipt < PTs.size(); ipt++){
		pixValue[0] = ip[0].getPixelValue((int) PTs[ipt].getX(), (int) PTs[ipt].getY());
		pixValue[1] = ip[1].getPixelValue((int) PTs[ipt].getX(), (int) PTs[ipt].getY());
		dataout.add(new String[]{ "" + PTs[ipt].getX() , "" + PTs[ipt].getY() ,
				"" + PTs[ipt].getX()*pixelSize[0] , "" + PTs[ipt].getY()*pixelSize[1],
				"" + pixValue[0], "" + pixValue[1],  "" +  area_convex, "" +  area_mask,
				"" +  PTs.size()/area_convex, "" +  PTs.size()/area_mask, "" + bestslice});
	}
	header = new String[]{"X_px", "Y_px", "X_um", "Y_um", "val_LoG",
			"val_raw", "area_convex_um2", "area_mask_um2", "density_convex_um2", "density_mask_um2", "bestSlice"};
	writeDataAtOnce(outfile, header, dataout, false);
	return PTs.size();
}



def writeDataAtOnce(outfile, header,   data,  append)  {
	// first create file object for file placed at location
	// specified by filepath
	//File file = new File(filePath);
	try {
		if (!outfile.exists())
			append = false;
		FileWriter outputfile = new FileWriter(outfile, append);
		// create CSVWriter object filewriter object as parameter
		CSVWriter writer = new CSVWriter(outputfile);
		if (!append) {
			writer.writeNext(header);
		}
		writer.writeAll(data);
		writer.close();
	}
	catch (IOException e) {
		e.printStackTrace();
	}
}

def checkPackages() {
	try {
		Class.forName("imagescience.image.Image");
		return true;
	} catch (Exception e) {
		IJ.showMessage("Please enable the ImageScience update site! [Help > Update > Manage update site > [x] ImageScience] ")
		return false;
	}
}